![ONNX Logo](vignettes/imgs/ONNX_logo_main.png) ![Go Logo](vignettes/imgs/Go-Logo_Blue.png)

[![](https://godoc.org/github.com/owulveryck/onnx-go?status.svg)](http://godoc.org/github.com/owulveryck/onnx-go)

This is a Go Interface to [Open Neural Network Exchange (ONNX)](https://onnx.ai/).

# Disclaimer

The implementation of the spec is partial on the import, and non-existent for the export.
**Warning** The API is experimental and may change.

# About

The target of this repository is to create an abstraction to create a conputation graph based on [the spec of ONNX](https://github.com/onnx/onnx/blob/master/docs/IR.md).

_for more information on computation graphs, please read this [blog post](http://gopherdata.io/post/deeplearning_in_go_part_1/)_

This package contains some facilities for reading binary version of an ONNX model and for unmarshaling it into a pointer to 
an object compatible with [`gonum's DirectedWeightedBuilder`](https://godoc.org/gonum.org/v1/gonum/graph#DirectedWeightedBuilder) interface.
The weigth of the edges is used to garantee the order of execution of the operands for non-commutative operations.

The entry point of this library is the `Unmarshal` function:

```go 
func Unmarshal(data []byte, dst graph.DirectedWeightedBuilder) error
```

## protobuf definition of ONNX

The protobuf definition of onnx has been converted into Go with the classic `protoc` tool. The definition can be found in the `internal` directory.
The definition is not exposed to avoid external dependencies to this repo. Indeed, the definition could change to use a more efficient compiler such
as `gogo protobuf` and this change should be transparent to the user of this package.

## Execution

For the graph to be executed, its node must carry references to tensors and operators.

### Tensors

onnx-go relies on the [`tensor package`](https://godoc.org/gorgonia.org/tensor) from the Gorgonia library.

If the node generated by the `NewNode` method of the interface is compatible with the [`TensorCarrier`](https://godoc.org/github.com/owulveryck/onnx-go#TensorCarrier)
interface, the method is called for each node referenced in the
[`Initializer`](https://github.com/onnx/onnx/blob/master/docs/IR.md) definition of ONNX (this includes the weigts of a pre-trained graph).

### Operators

If the graph builder is compatible with the [`OperationCarrier`](https://godoc.org/github.com/owulveryck/onnx-go#OperationCarrier) interface, the corresponding method is applied on each node referenced in the Node array of the graph in the onnx definition.

## Backend

onnx-go do not provide any executable backend, but for a reference, a simple backend that builds an information graph is provided as an example (see the `simple` subpackage).

### Gorgonia

A tweaked version of Gorgonia is vendored in the example directory. Next official release of Gorgonia should be compatible with the ONNX interface out-of-the box. Meanwhile, feel free to play
with the vendorer version that is delivered as-is without any guarantee.

# Example: MNIST

The vendored version of Gorgonia can run the MNIST model.

```sh
curl https://www.cntk.ai/OnnxModels/mnist/opset_7/mnist.tar.gz | tar -C /tmp -xzvf -
```

Example:

```go
import (
        "log"

        onnx "github.com/owulveryck/onnx-go"
        "github.com/owulveryck/onnx-go/internal/examples/mnist"
        pb "github.com/owulveryck/onnx-go/internal/pb-onnx"
        "gorgonia.org/gorgonia/debugger/dot"
        "gorgonia.org/gorgonia/node"
        gorgonnx "gorgonia.org/gorgonia/onnx"
)
//...
        // Create a graph compatible with gonum WeightedDirectedBuilder
        graph := gorgonnx.NewGraph()
        // Create the model based on this graph
        m := onnx.NewModel(graph)
        // Get the binary of the model
        b, err := ioutil.ReadFile("/tmp/mnist/model.onnx")
        // ...
        err := m.Unmarshal(b)
        // Draw a representation of the graph in dot
        dotB, err = dot.Marshal(graph)

        // ... 
        sampleTestData := new(pb.TensorProto)
        // Read a sample file
        b, err = ioutil.ReadFile("/tmp/mnist/test_data_set_0/input_0.pb")
        //err = sampleTestData.XXX_Unmarshal(mnist.GetTest1Input0())
        err = sampleTestData.XXX_Unmarshal(b)
        
        // Convert the data to a tensor.Tensor
        t, err := sampleTestData.Tensor()

        // Set the input value of the graph
        err = gorgonnx.Let(graph.Node(m.Input[0]).(node.Node), t)

        // create a VM to run the program on
        machine := gorgonnx.NewTapeMachine(graph)

        // Run the program
        err = machine.RunAll()
        if err != nil {
              log.Fatal(err)
        }
        for _, v := range m.Output {
              // Print the result
              log.Println(graph.Node(v).(node.Node).Value().Data())
        }
```
